
/* Copyright (c) 1985, 1989 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by the University of
 *  California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* 
 * From: @(#)ftp.c  5.38 (Berkeley) 4/22/91
 */

/**
 * @addtogroup ftp
 * @{
 **/

/**
 * @file ftp.c
 * @brief Implements File Transfer Protocol
 *
 * @author Copyright (c) 1985, 1989 Regents of the University of California.All rights reserved.
 *
 * @note From: @(#)ftp.c  5.38 (Berkeley) 4/22/91
 *
 * */

#include <types.h>
#include "stdlib.h"

#include <sys/socket.h>
#include <sys/file.h>

#include <sys/select.h>

#include <arpa/inet.h>

#include <unistd.h>
#include <errno.h>

#include <log_pub.h>
#include <ctype.h>

#include "ftp_var.h"
#include "version.h"
#include "telnet_constants.h"

int             data = -1;

static struct sockaddr_in hisctladdr;
static struct sockaddr_in data_addr;
static struct sockaddr_in myctladdr;
static int      abrtflag = 0;

/// Variable defined in ftp_var.h
int             connected;
int             code;
int             crflag;
char            pasv[64];       ///< passive port for proxy data connection
int             type;
int             curtype;
char            bytename[32];   ///<  local byte size in ascii */
char           *hostname;       ///< name of host connected to */
int             cpend;          ///< flag: if != 0, then pending server reply */
int             options;        ///< used during socket creation */
char            reply_string[BUFSIZ]; ///< last line of previous reply */
char            lreply[BUFSIZ];

/// Extern variables defined in ftplib.h
extern int      ftp_passive;
extern int      ftp_verbose;
extern int      ftp_sendport;
extern int      ftp_debug;

/*
 * static void abort_remote(int data_fd);
 * static int ptabflg = 0;
 */

int             control_fd = -1;
int             data_socket_error = 0;

struct types    tipos[] = {
  {"ascii", "A", TYPE_A, NULL},
  {"binary", "I", TYPE_I, NULL},
  {"image", "I", TYPE_I, NULL},
  {"ebcdic", "E", TYPE_E, NULL},
  {"tenex", "L", TYPE_L, bytename},
  {NULL, NULL, 0, NULL}
};

static t_log    ftpLog = { "ftp", LOG_LEVEL_ERROR | LOG_LEVEL_WARNING };

char           *hookup(char *host, int port)
{
#ifdef IP_TOS
  int             tos;
#endif
  socklen_t       len;
  static char     hostnamebuf[256];
  int             iRc;
  int             iSockFlags;
  struct timeval  L_timeout;
  fd_set          L_wset;

  memset(&hisctladdr, 0, sizeof(struct sockaddr_in));
  if (inet_aton(host, &hisctladdr.sin_addr))
  {
    hisctladdr.sin_family = AF_INET;
    hisctladdr.sin_port = htons(port);
    strncpy(hostnamebuf, host, sizeof(hostnamebuf));
    hostnamebuf[sizeof(hostnamebuf) - 1] = 0;
  }
  else
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::hookup IP dot format required\n");
    code = -1;
    return (NULL);
  }
  hostname = hostnamebuf;
  control_fd = socket(hisctladdr.sin_family, SOCK_STREAM, 0);
  if (control_fd < 0)
  {
    LogPrint(ftpLogGet(), 0, "Ftp::hookup socket\n");
    code = -1;
    close(control_fd);
    control_fd = -1;
    return (NULL);
  }

  iSockFlags = fcntl(control_fd, F_GETFL, 0);
  iSockFlags |= O_NONBLOCK;
  iRc = fcntl(control_fd, F_SETFL, iSockFlags);

  if (iRc < 0)
  {
    code = -1;
    close(control_fd);
    control_fd = -1;
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::hookup: Can't set NONBLOCK socket\n");
    return (NULL);
  }
  if (connect(control_fd, (struct sockaddr *)&hisctladdr, sizeof(hisctladdr)) < 0)
  {
    if (errno != EINPROGRESS)
    {
      LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::hookup connect failure");
      code = -1;
      close(control_fd);
      control_fd = -1;
      return (NULL);
    }
    L_timeout.tv_sec = CTRL_TIMEOUT;
    L_timeout.tv_usec = 0;
    FD_ZERO(&L_wset);
    FD_SET(control_fd, &L_wset);
    if (select(control_fd + 1, NULL, &L_wset, NULL, &L_timeout) <= 0)
    {
      LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::hookup connect failure");
      code = -1;
      close(control_fd);
      control_fd = -1;
      return (0);
    }
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::hookup connected");
  }
  else
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "****************** ERROR 1");
  }

  len = sizeof(struct sockaddr_in);
  if (getsockname(control_fd, (struct sockaddr *)&myctladdr, &len) < 0)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::hookup getsockname");
    code = -1;
    close(control_fd);
    control_fd = -1;
    return (NULL);
  }

#ifdef IP_TOS
  tos = IPTOS_LOWDELAY;
  if (setsockopt(control_fd, IPPROTO_IP, IP_TOS, (char *)&tos, sizeof(int)) < 0)
    LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::hookup setsockopt TOS (ignored)");
#endif
  if (getreply(0) > 2)
  {                             /* read startup message from server */
    code = -1;
    close(control_fd);
    control_fd = -1;
    return (NULL);
  }
  LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::hookup Have the startup message from server!\n");
#ifdef SO_OOBINLINE
  {
    int             on = 1;

    if (setsockopt(control_fd, SOL_SOCKET, SO_OOBINLINE, (char *)&on, sizeof(on)) < 0 && ftp_debug)
    {
      LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::hookup setsockopt");
    }
  }
#endif /* SO_OOBINLINE */

  return (hostname);
}

int dologin(const char *luser, const char *pass)
{
  int             n;
  char            cmd[MAX_BUF];

  strncpy(cmd, "USER ", MAX_BUF);
  strncpy(cmd + 5, luser, MAX_BUF - 5);
  n = command(cmd);
  if (n == FTP_CONTINUE)
  {
    strncpy(cmd, "PASS ", MAX_BUF);
    strncpy(cmd + 5, pass, MAX_BUF - 5);
    n = command(cmd);
  }
  if (n == FTP_CONTINUE)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::dologin Server asked account!!");
    return 1;
  }
  if (n != FTP_COMPLETE)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::dologin Login failed.");
    return 1;
  }
  return 0;
}

int command(const char *cmd)
{
  int             r;
  int             sent = 0;
  int             L_cmdLength = 0;
  int             ret = -1;
  struct timeval  L_timeout;
  fd_set          L_wset;
  char            buf[MAX_BUF];

  abrtflag = 0;
  if (ftp_debug)
  {
    if (strncmp("PASS ", cmd, 5) == 0)
      LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "PASS XXXX");
    else
      LogPrint(ftpLogGet(), LOG_LEVEL_INFO, cmd);
  }
  if (control_fd == -1)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::command No control connection for command");
    code = -1;
    return (0);
  }
  L_timeout.tv_sec = CTRL_TIMEOUT;
  L_timeout.tv_usec = 0;
  FD_ZERO(&L_wset);
  FD_SET(control_fd, &L_wset);
  snprintf(buf, MAX_BUF, "%s\r\n", cmd);
  LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::command: %s\n", buf);
  if ((ret = select(control_fd + 1, NULL, &L_wset, NULL, &L_timeout)) > 0)
  {
    L_cmdLength = strlen(buf);
    if ((sent = write(control_fd, buf, L_cmdLength)) != L_cmdLength)
    {
      if (errno == ETIMEDOUT)
        LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Connection time-out. You have to connect again");
      else
        LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::command Write error. Sent: %d (%d) errno:%d",
                 sent, L_cmdLength, errno);
      lostpeer(0);
      code = -1;
      return (0);
    }
    cpend = 1;
    r = getreply(!strcmp(cmd, "QUIT"));
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Getreply result:%d", r);
    return (r);
  }
  return 0;
}

int read_char(int fd)
{
  unsigned char   mychar = 0;
  int             L_read = 0;

  /* local variables for non-blocking mode */
  fd_set          L_rset;
  struct timeval  L_timeout;
  int             ret = -1;
  int             loop = 1;

  /* non-blocking implementation */
  L_timeout.tv_sec = DATA_TIMEOUT;
  L_timeout.tv_usec = 0;
  while (loop)
  {
    loop = 0;
    FD_ZERO(&L_rset);
    FD_SET(fd, &L_rset);
    if ((ret = select(fd + 1, &L_rset, NULL, NULL, &L_timeout)) == 1)
    {
      if ((L_read = recv(fd, &mychar, 1, 0)) == 1)
      {
        return mychar;
      }
      else if (L_read < 0)
      {
        if (errno == ESHUTDOWN)
        {
          LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "read_char: connection closed by peer");
          return EOF;
        }
        else if (errno == EAGAIN)
        {
          //LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "read_char: EAGAIN error ");
          /* redo the select loop */
          loop = 1;
        }
        else
          LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "read_char: read error - errno %d", errno);
      }
    }
    else if (ret == -1)
    {
      LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "read_char: select error - errno %d", errno);
    }
    else if (ret == 0)
    {
      /* timeout expired: must not return an error, since
       * it can be a "normal connection closing 
       * Indeed, we are currently unable to detect 
       * the graceful connection closign by peer
       * when it is done with both FIN and PSH flags */
      LogPrint(ftpLogGet(), LOG_LEVEL_WARNING, "read_char: no data to read for %d seconds",
               L_timeout.tv_sec);
      return EOF;
    }
  }
  data_socket_error = 1;
  return EOF;
}

#if 0
int put_char(int fd, int ch)
{
  unsigned char   mychar = 0;

  /* local variables for non-blocking mode */
  fd_set          L_wset;
  struct timeval  L_timeout;
  int             ret = -1, L_read = -1;

  mychar = ch;
  /* blocking implementation
   * if (write(fd, &mychar, 1) != 1){
   * LogPrint (, LOG_LEVEL_INFO, "Ftp::put_char write error");
   * return -1;
   * }
   * return 0;
   */

  /* non-blocking implementation */
  L_timeout.tv_sec = DATA_TIMEOUT;
  L_timeout.tv_usec = 0;
  FD_ZERO(&L_wset);
  FD_SET(fd, &L_wset);
  if ((ret = select(fd + 1, NULL, &L_wset, NULL, &L_timeout)) == 1)
  {
    if ((L_read = write(fd, &mychar, 1)) == 1)
      return 0;
    else
      LogPrint(, LOG_LEVEL_INFO, "Ftp::put_char write error");
  }
  else if (ret == -1)
  {
    LogPrint(, LOG_LEVEL_INFO, "Ftp::put_char select error %d", errno);
  }
  else if (ret == 0)
  {
    /* timeout expired: consider this as a failure case */
    LogPrint(, LOG_LEVEL_INFO, "Ftp::put_char select timeout");
  }
  return -1;
}
#endif

int getreply(int expecteof)
{
  register int    c, n;
  register int    dig;
  register char  *cp, *lcp;
  int             originalcode = 0, continuation = 0;

  //  void (*oldintr)(int);
  int             pflag = 0;
  size_t          px = 0;
  size_t          psize = sizeof(pasv);
  char            send_buffer[4];
  int             sent = 0;

  lcp = lreply;
  //  oldintr = signal(SIGINT, cmdabort);
  for (;;)
  {
    dig = n = code = 0;
    cp = reply_string;
    while ((c = read_char(control_fd)) != '\n')
    {
      if (c == IAC)
      {                         /* handle telnet commands */
        switch (c = read_char(control_fd))
        {
          case WILL:
          case WONT:
            c = read_char(control_fd);
            sprintf(send_buffer, "%c%c%c", IAC, DONT, c);
            if ((sent = write(control_fd, send_buffer, 3)) != 3)
            {
              LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::getreply write error");
            }
            break;
          case DO:
          case DONT:
            c = read_char(control_fd);
            sprintf(send_buffer, "%c%c%c", IAC, WONT, c);
            if ((sent = write(control_fd, send_buffer, 3)) != 3)
            {
              LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::getreply write error");
            }
            break;
          default:
            break;
        }
        continue;
      }
      dig++;
      if (c == EOF)
      {
        if (expecteof)
        {
          code = 221;
          return (0);
        }
        lostpeer(0);
        if (ftp_verbose)
        {
          LogPrint(ftpLogGet(), LOG_LEVEL_ERROR,
                   "421 Service not available, remote server has closed connection");
          fflush(stdout);
        }
        code = 421;
        return (4);
      }
      /*
       * if (c != '\r' && (ftp_verbose > 0 ||
       * (ftp_verbose > -1 && n == '5' && dig > 4))) {
       * (void) putchar(c);
       * }
       */
      if (dig < 4 && isdigit(c))
        code = code * 10 + (c - '0');
      if (!pflag && code == 227)
        pflag = 1;
      if (dig > 4 && pflag == 1 && isdigit(c))
        pflag = 2;
      if (pflag == 2)
      {
        if (c != '\r' && c != ')')
        {
          if (px < psize - 1)
            pasv[px++] = c;
        }
        else
        {
          pasv[px] = '\0';
          pflag = 3;
        }
      }
      if (dig == 4 && c == '-')
      {
        if (continuation)
          code = 0;
        continuation++;
      }
      if (n == 0)
        n = c;

      if (c != '\r' && (lcp - lreply < BUFSIZ - 1))
        *lcp++ = c;
      if (cp < &reply_string[sizeof(reply_string) - 1])
        *cp++ = c;
    }
    if (lcp - lreply < BUFSIZ - 1)
      *lcp++ = c;
    /*
     * if (ftp_verbose > 0 || (ftp_verbose > -1 && n == '5')) {
     * (void) putchar(c);
     * (void) fflush (stdout);
     * }
     */
    if (continuation && code != originalcode)
    {
      if (originalcode == 0)
        originalcode = code;
      continue;
    }
    *cp = '\0';
    *lcp = '\0';
    if (n != '1')
      cpend = 0;
    if (code == 421 || originalcode == 421 || code == 425 || originalcode == 425)
      lostpeer(0);

    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Getreply. code:%d originalcode:%d \n", code,
             originalcode);

    if ((code == 550 || originalcode == 550) || (code == 552 || originalcode == 552) ||
        (code == 553 || originalcode == 553) || (code == 450 || originalcode == 450) || (code == 452
                                                                                         ||
                                                                                         originalcode
                                                                                         == 452))
      return (FTP_FILE_ERROR);
    return (n - '0');
  }
}

void changetype(int newtype, int show)
{
  struct types   *p;
  int             comret, oldverbose = ftp_verbose;
  char            cmd[MAX_BUF];
  int             len;

  if (newtype == 0)
    newtype = TYPE_I;
  if (newtype == curtype)
    return;
  if (ftp_debug == 0 && show == 0)
    ftp_verbose = 0;
  for (p = tipos; p->t_name; p++)
    if (newtype == p->t_type)
      break;
  if (p->t_name == 0)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::changetype Internal error: unknown type %d",
             newtype);
    return;
  }
  if (newtype == TYPE_L && bytename[0] != '\0')
  {
    strncpy(cmd, "TYPE ", MAX_BUF);
    strncpy(cmd + 5, p->t_mode, MAX_BUF - 5);
    len = strlen(cmd);
    strncpy(cmd + len, bytename, MAX_BUF - len);
    comret = command(cmd);
  }
  else
  {
    strncpy(cmd, "TYPE ", MAX_BUF);
    strncpy(cmd + 5, p->t_mode, MAX_BUF - 5);
    comret = command(cmd);
  }
  if (comret == FTP_COMPLETE)
    curtype = newtype;
  ftp_verbose = oldverbose;
}

int sendrequest(const char *cmd, char *remote, int (*in_fct) (char *, unsigned), unsigned bufsize,
                off_t restart_point)
{
  struct timeval  start, stop;
  register int    c = 0, d = 0;
  volatile long   bytes = 0;

  char           *buf = NULL;
  char            cmd_str[MAX_BUF];
  int             lcmd_str;
  char           *bufp;

  /* local variables for non-blocking mode */
  fd_set          L_wset;
  struct timeval  L_timeout;
  int             send_retries = 0;
  int             ret = -1;

  if ((buf = malloc(bufsize)) == NULL)
  {
    return (1);
  }

  if (curtype != type)
    changetype(type, 0);

  if (initconn())
  {
    code = -1;
    free(buf);
    return 1;
  }

  if (restart_point)
  {
    snprintf(cmd_str, MAX_BUF, "REST %ld", (long)restart_point);
    if (command(cmd_str) != FTP_CONTINUE)
    {
      free(buf);
      return 1;
    }
  }

  strncpy(cmd_str, cmd, MAX_BUF);
  lcmd_str = strlen(cmd_str);
  strncpy(cmd_str + lcmd_str, remote, MAX_BUF - lcmd_str);
  if (command(cmd_str) != FTP_PRELIM)
  {
    free(buf);
    return 1;
  }

  if (dataconn())
    goto abort;

  gettimeofday(&start, (struct timezone *)0);
  LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "ftp:sendrequest: command:%s curtype:%d\n", cmd, curtype);
  switch (curtype)
  {

    case TYPE_I:
    case TYPE_L:
      /* Blocking mode implementation
       * Cannot be used since setsockopt(SO_RCVTIMEO) option
       * is not supported 
       *
       * errno = d = 0;
       * while ((c = (*in_fct)(buf, bufsize)) > 0) {
       *  bytes += c;
       *  for (bufp = buf; c > 0; c -= d, bufp += d)
       *    if ((d = write(data, bufp, c)) <= 0)
       *      break;
       * }
       * if (c < 0)
       *  LogPrint(, LOG_LEVEL_INFO, "Ftp::sendrequest Out callback returned %d", c);
       * if (d < 0) {
       *  if (errno != EPIPE) 
       *    LogPrint(, LOG_LEVEL_INFO, "Ftp::sendrequest netout");
       *  bytes = -1;
       * }
       * break;
       */

      /* Non-blocking implementation */
      errno = d = 0;
      LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "ftp:sendrequest: TYPE_I/L\n");
      L_timeout.tv_sec = DATA_TIMEOUT;
      L_timeout.tv_usec = 0;
      while ((c = (*in_fct) (buf, bufsize)) > 0)
      {
        bytes += c;
        send_retries = 0;
        for (bufp = buf; c > 0; c -= d, bufp += d)
        {
          FD_ZERO(&L_wset);
          FD_SET(data, &L_wset);
          if ((ret = select(data + 1, NULL, &L_wset, NULL, &L_timeout)) == 1)
          {
            if ((d = write(data, bufp, c)) <= 0)
            {
              LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Write failed after chunk c:%d: errno %d\n", c,
                       errno);
              goto abort;
            }
          }
          else if (ret == 0)
          {
            /* Timeout */
            LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Timeout: cannot send data for %d seconds",
                     L_timeout.tv_sec);
            /* since we are not able to detect closing connection
             * in all cases (in particular within a TCP packet that
             * has both FIN and PSH flags), we have to return 
             * success 
             break; */
            if (send_retries++ == MAX_WRITE_RETRIES)
              goto abort;
          }
          else if (ret == -1)
          {
            LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::sendrequest select error %d", errno);
            bytes = -1;
            goto abort;
          }
        }
      }
      if (c < 0)
        LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::sendrequest Out callback returned %d", c);
      if (d < 0)
      {
        LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::sendrequest write error %d", errno);
        bytes = -1;
      }
      break;

    case TYPE_A:
      /*
       * while ((l=(*in_fct)(buf, bufsize))) {
       * for (cpt=0; cpt<l; cpt++) {
       * c = buf[cpt];
       * if (c == '\n') {
       * if (put_char(data, '\r'))
       * break;
       * bytes++;
       * }
       * put_char(data, c);
       * bytes++;
       * }
       * }
       */
      LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "ftp:sendrequest: TYPE_A\n");
      LogPrint(ftpLogGet(), LOG_LEVEL_ERROR,
               "Ftp::sendrequest Text mode not supported for transfer!!!");
      bytes = -1;
      break;
  }
  gettimeofday(&stop, (struct timezone *)0);
  close(data);
  data = -1;
  if (getreply(0) != FTP_COMPLETE)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::sendrequest Transfer not positively completed");
    bytes = -1;
  }
  if (bytes > 0)
  {
    ptransfer("sent", bytes, &start, &stop);
  }

  if (buf != NULL)
  {
    free(buf);
    buf = NULL;
  }

  return (bytes < 0);
abort:
  gettimeofday(&stop, (struct timezone *)0);

  if (buf != NULL)
  {
    free(buf);
    buf = NULL;
  }

  if (!cpend)
  {
    code = -1;
    return 1;
  }
  if (data >= 0)
  {
    close(data);
    data = -1;
  }
  getreply(0);
  code = -1;
  if (bytes > 0)
  {
    ptransfer("sent", bytes, &start, &stop);
  }

  return 1;
}

int recvrequest(const char *cmd, char *remote, int (*out_fct) (char *, unsigned), unsigned bufsize,
                off_t restart_point)
{
  volatile int    is_retr, tcrflag, bare_lfs = 0;
  volatile long   bytes = 0;
  register int    c = 0;
  struct timeval  start, stop;

  static char    *buf = NULL;
  static unsigned cpt = 0;

  /* local variables for non-blocking mode */
  struct timespec sleepTime;
  int             L_internalError = 0;
  int             L_dataFinished = 0;
  char            cmd_str[MAX_BUF];
  int             lcmd_str;
  int             err;

  if ((buf = malloc(bufsize)) == NULL)
  {
    return (1);
  }

  is_retr = strncmp(cmd, "RETR", 4) == 0;
  tcrflag = !crflag && is_retr;

  LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "is_retr:%d curtype:%d type:%d cmd:%s\n", is_retr, curtype,
           type, cmd);
  if (!is_retr)
  {
    if (curtype != TYPE_A)
    {
      changetype(TYPE_A, 0);
    }
  }
  else if (curtype != type)
  {
    changetype(type, 0);
  }
  // Start data connection
  if (initconn())
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "ftp:recvrequest: can't start data connection\n",
             command);
    code = -1;
    free(buf);
    return 1;
  }
  LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "ftp:recvrequest: data connection stablished\n");
  if (restart_point)
  {
    snprintf(cmd_str, MAX_BUF, "REST %ld", (long)restart_point);
    if (is_retr && restart_point && (err = command(cmd_str)) != FTP_CONTINUE)
    {
      LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "ftp:recvrequest: err:%d\n", err);
      free(buf);
      return err;
    }
  }
  if (remote)
  {
    strncpy(cmd_str, cmd, MAX_BUF);
    lcmd_str = strlen(cmd_str);
    strncpy(cmd_str + lcmd_str, remote, MAX_BUF - lcmd_str);
    if ((err = command(cmd_str)) != FTP_PRELIM)
    {
      LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "ftp:recvrequest: err:%d\n", err);
      free(buf);
      return err;
    }
  }
  else
  {
    if ((err = command(cmd)) != FTP_PRELIM)
    {
      LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "ftp:recvrequest: err:%d\n", err);
      free(buf);
      return err;
    }
  }
  LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "ftp:recvrequest: command:%s curtype:%d\n", cmd, curtype);
  if (dataconn())
    goto abort;

  gettimeofday(&start, (struct timezone *)0);

  switch (curtype)
  {

    case TYPE_I:
    case TYPE_L:
      /* Blocking mode implementation
       * Cannot be used since setsockopt(SO_RCVTIMEO) option
       * is not supported 
       *
       * errno = 0;
       * while ((c = read(data , buf, bufsize)) > 0) {
       *  if ((*out_fct)(buf,c)) goto abort;
       *  bytes += c;
       * }
       * if  (c < 0) {
       *  if (errno == ESHUTDOWN){
       *    LogPrint(, LOG_LEVEL_INFO, "Ftp::recvrequest Connection closed by peer");
       *    break;
       *  }
       *  if (errno != EPIPE)
       *    LogPrint(, LOG_LEVEL_INFO, "netin");
       *  bytes = -1;
       * }
       * break;
       */

      /* Non-blocking implementation */

      LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "ftp:recvrequest: TYPE_I/L\n");

      do
      {
        memset(buf, 0, bufsize);
        c = recv(data, buf, bufsize, 0);

        if (c == -1)
        {
          if (errno == ESHUTDOWN)
          {
            //NOTE: Netergy's tcp implementation works like this: When the other end closes
            //      the socket, recv() returns (-1) and errno=ESHUTDOWN instead of the MORE
            //      COMMON AND EXPECTED (Unix-like) behaviour, i.e. recv() returns 0.
            //      So... we must treat this case the same as a normal socket closing
            //      operation.
            LogPrint(ftpLogGet(), LOG_LEVEL_WARNING,
                     "Ftp::recvrequest Connection closed by peer (read() == -1)\n");
            L_dataFinished = 1;
            break;              // exit the do-while{} loop
          }
          else
          {
            // This is a REAL error.
            LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::recvrequest recv() error. Errno = %d\n",
                     errno);
            bytes = -1;
            break;              // exit the do-while{} loop
          }
        }
        else if (c == 0)
        {
          // This is how the connection SHOULD end, however Netergy's TCP implementation
          // returns "0" when the socket is NON-BLOCKING and there is not data to read...
          // instead of the MORE COMMON AND EXPECTED (Unix-like) behaviour, i.e. recv()
          // returns 0 when the other end closes the socket.
          // So this will never happen... However we protect this here to be more "portable", (ehem)
          LogPrint(ftpLogGet(), LOG_LEVEL_INFO,
                   "Ftp::recvrequest Connection closed by peer (read() == 0)\n");
          L_dataFinished = 1;
          break;                // exit the do-while{} loop
        }
        else if (c > 0)
        {
          // We have received data from the ftp-data socket. Proccess it.
          if ((*out_fct) (buf, c))
          {
            // internal error
            L_internalError = 1;
            LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "ftp:Internal error (local error)");
            break;
          }
          else
          {
            bytes += c;
          }
        }
        else
        {
          // c < -1 ???? ERROR !
          L_internalError = 1;
          bytes = -1;
          LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "ftp:Internal error (recv() < -1 ???)");
        }

      }
      while (c != 0);

      if (L_internalError)
        goto abort;

      break;

    case TYPE_A:
      /*
       * LogPrint(, LOG_LEVEL_INFO, "Ftp::recvrequest Text mode not supported for transfer!!!");
       * bytes = -1;
       */

      LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "ftp:recvrequest: TYPE_A\n");
      while ((c = read_char(data)) != EOF)
      {
        if (c == '\n')
          bare_lfs++;
        while (c == '\r')
        {
          bytes++;
          if ((c = read_char(data)) != '\n' || tcrflag)
          {
            if (c == '\0')
            {
              bytes++;
              goto contin2;
            }
            if (c == EOF)
              goto contin2;
          }
        }
        if (cpt == (bufsize - 1))
        {
          buf[cpt] = c;
          buf[cpt++] = 0;
          if ((*out_fct) (buf, cpt))
            goto abort;
          cpt = 0;
        }
        // Es discutible si en modo ASCII debe quitar los \n
        if (c == '\n')
        {
          buf[cpt] = 0;
          if ((*out_fct) (buf, cpt))
            goto abort;
          cpt = 0;
        }
        else
          buf[cpt++] = c;
        bytes++;
      contin2:;
      }
      if (bare_lfs)
      {
        LogPrint(ftpLogGet(), LOG_LEVEL_WARNING,
                 "WARNING! %d bare linefeeds received in ASCII mode", bare_lfs);
        LogPrint(ftpLogGet(), LOG_LEVEL_WARNING, "File may not have transferred correctly.");
      }
      if (data_socket_error)
      {
        LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::recvrequest data_socket_error");
        data_socket_error = 0;
        bytes = -1;
      }

      break;
  }

  gettimeofday(&stop, (struct timezone *)0);

  sleepTime.tv_sec = 1;
  sleepTime.tv_nsec = 0;
  nanosleep(&sleepTime, NULL);

  // Now we write "ghost" data. This is
  // a dirty trick to force some ftp
  // servers (particularly Linux proftpd)
  // to completely close the data socket,
  // so that a new "flup" operation on
  // the same port can be performed
  // right away.
  write(data, "4815162342", 10);
  nanosleep(&sleepTime, NULL);

  shutdown(data, SHUT_RDWR);
  close(data);
  data = -1;

  LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::recvrequest Checking data 'completeness'");

  // In case the ftp-data channel was clossed by the other end, we guess data
  // has completely been tx'ed. This happens when L_dataFinished == 0.
  // In any other case we check the ftp-control channel to see if the 226
  // message (FTP tx ok) was transmitted... not every ftp server sends
  // this message, and some of them send this message as soon as the tx
  // starts, before knowing if everything went ok... So this code is
  // near to useless.

  if (L_dataFinished == 0)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_WARNING,
             "Ftp::recvrequest Was data completely tx'ed? Checking ftp-control channel");

    if (getreply(0) != FTP_COMPLETE)
    {
      LogPrint(ftpLogGet(), LOG_LEVEL_WARNING, "Ftp::recvrequest Incompleted data !!!");
      bytes = -1;
    }
    else
    {
      LogPrint(ftpLogGet(), LOG_LEVEL_INFO,
               "Ftp::recvrequest FTP_COMPLETE message received. Guessing data was tx'ed ok.");
    }
  }
  else
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO,
             "Ftp::recvrequest Peer closed ftp-data channel. Guessing data was tx'ed ok.");
  }

  if (bytes > 0 && is_retr)
  {
    ptransfer("received", bytes, &start, &stop);
  }

  if (buf != NULL)
  {
    free(buf);
    buf = NULL;
  }

  return (bytes < 0);
abort:

  /* abort using RFC959 recommended IP,SYNC sequence  */
  gettimeofday(&stop, (struct timezone *)0);

  if (buf != NULL)
  {
    free(buf);
    buf = NULL;
  }

  if (!cpend)
  {
    code = -1;
    return 1;
  }
  code = -1;
  lostpeer(0);

  /*
   * Actually, the standard cannot be implemented since OOB mode is not 
   * supported in the current TCP Netergy stack
   * In place of abort_remote invocation, we close the control connection
   */
#if 0
  abort_remote(data);
  code = -1;
  if (data >= 0)
  {
    shutdown(data, SHUT_RDWR);
    close(data);
    data = -1;
  }
  if (control_fd >= 0)
  {
    (void)close(control_fd);
    control_fd = -1;
    connected = 0;
  }
#endif
  if (bytes > 0)
    ptransfer("received", bytes, &start, &stop);

  return 1;
}

/*
 * Need to start a listen on the data channel before we send the command,
 * otherwise the server's connect may fail.
 */
int initconn(void)
{
  register char  *p, *a;
  int             result, tmpno = 0;
  socklen_t       len;
  int             on = 1;

  char            cmd[MAX_BUF];

  /*
   * #ifdef IP_TOS
   * //    int tos;
   * #endif
   * 
   * unsigned long a1,a2,a3,a4,p1,p2;
   * 
   * if (ftp_passive) {
   * data = socket(AF_INET, SOCK_STREAM, 0);
   * if (data < 0) {
   * LogPrint(, LOG_LEVEL_INFO, "Ftp::initconn socket");
   * return(1);
   * }
   * if (options & SO_DEBUG &&
   * setsockopt(data, SOL_SOCKET, SO_DEBUG, (char *)&on,
   * sizeof (on)) < 0){
   * close(data);
   * data = -1;
   * LogPrint(, LOG_LEVEL_INFO, "Ftp::initconn setsockopt (ignored)");
   * }
   * if (command("PASV") != FTP_COMPLETE) {
   * LogPrint(, LOG_LEVEL_INFO, "Ftp::initconn Passive mode refused.");
   * close(data);
   * data = -1;
   * return(1);
   * }
   */

  /*
   * What we've got at this point is a string of comma separated
   * one-byte unsigned integer values, separated by commas.
   * The first four are the an IP address. The fifth is the MSB
   * of the port number, the sixth is the LSB. From that we'll
   * prepare a sockaddr_in.
   */

  /*
   * if (sscanf(pasv,"%lu,%lu,%lu,%lu,%lu,%lu",
   * &a1,&a2,&a3,&a4,&p1,&p2)
   * != 6) 
   * {
   * LogPrint(, LOG_LEVEL_INFO, "Ftp::initconn Passive mode address scan failure. Shouldn't happen!");
   * close(data);
   * data = -1;
   * return(1);
   * }
   * 
   * data_addr.sin_family = AF_INET;
   * data_addr.sin_addr.s_addr = htonl((a1 << 24) | (a2 << 16) |
   * (a3 << 8) | a4);
   * data_addr.sin_port = htons((p1 << 8) | p2);
   * 
   * if (connect(data, (struct sockaddr *) &data_addr,
   * sizeof(data_addr))<0) {
   * LogPrint(, LOG_LEVEL_INFO, "Ftp::initconn connect");
   * close(data);
   * data = -1;
   * return(1);
   * }
   * #ifdef IP_TOS
   * tos = IPTOS_THROUGHPUT;
   * if (setsockopt(data, IPPROTO_IP, IP_TOS, (char *)&tos,
   * sizeof(tos)) < 0)
   * LogPrint(, LOG_LEVEL_INFO, "Ftp::initconn setsockopt TOS (ignored)");
   * #endif
   * return(0);
   * }
   */
noport:
  data_addr = myctladdr;
  if (ftp_sendport)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO,
             "Ftp::initconn Let system pick one port for the data connection");
    data_addr.sin_port = 0;     /* let system pick one */
  }
  if (data != -1)
    close(data);
  data = socket(AF_INET, SOCK_STREAM, 0);
  if (data < 0)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::initconn socket");
    if (tmpno)
      ftp_sendport = 1;
    return (1);
  }
  if (!ftp_sendport)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::initconn Allow SO_REUSEADDR");
    if (setsockopt(data, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) < 0)
    {
      LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::initconn setsockopt (reuse address)");
      goto bad;
    }
  }
  len = sizeof(data_addr);
  if (bind(data, (struct sockaddr *)&data_addr, len) < 0)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::initconn bind");
    goto bad;
  }
  if (options & SO_DEBUG && setsockopt(data, SOL_SOCKET, SO_DEBUG, (char *)&on, sizeof(on)) < 0)
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::initconn setsockopt (ignored)");
  if (getsockname(data, (struct sockaddr *)&data_addr, &len) < 0)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::initconn getsockname");
    goto bad;
  }
  if (listen(data, 1) < 0)
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::initconn listen");
  if (ftp_sendport)
  {
    a = (char *)&data_addr.sin_addr;
    p = (char *)&data_addr.sin_port;
#define UC(b) (((int)b)&0xff)
    snprintf(cmd, MAX_BUF, "PORT %d,%d,%d,%d,%d,%d", UC(a[0]), UC(a[1]), UC(a[2]), UC(a[3]),
             UC(p[0]), UC(p[1]));
    result = command(cmd);
    //        command("PORT %d,%d,%d,%d,%d,%d",
    //          UC(a[0]), UC(a[1]), UC(a[2]), UC(a[3]),
    //          UC(p[0]), UC(p[1]));
    if (result == FTP_ERROR && ftp_sendport == -1)
    {
      ftp_sendport = 0;
      tmpno = 1;
      goto noport;
    }
    return (result != FTP_COMPLETE);
  }
  if (tmpno)
    ftp_sendport = 1;
#ifdef IP_TOS
  on = IPTOS_THROUGHPUT;
  if (setsockopt(data, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int)) < 0)
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::initconn setsockopt TOS (ignored)\n");
#endif
  return (0);
bad:
  (void)close(data), data = -1;
  if (tmpno)
    ftp_sendport = 1;
  return (1);
}

int dataconn()
{
  struct sockaddr_in from;
  int             s;

#ifdef IP_TOS
  int             tos;
#endif
  socklen_t       fromlen = sizeof(from);

  //  int L_blockingMode = 1; /* non-blocking mode */

  if (ftp_passive)
    return 0;

  s = accept(data, (struct sockaddr *)&from, &fromlen);
  if (s < 0)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::dataconn accept error\n");
    close(data);
    data = -1;
    return 1;
  }
  close(data);
  data = s;
#ifdef IP_TOS
  tos = IPTOS_THROUGHPUT;
  if (setsockopt(s, IPPROTO_IP, IP_TOS, (char *)&tos, sizeof(int)) < 0)
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::dataconn setsockopt TOS (ignored)\n");
#endif
  return 0;
}

void ptransfer(const char *direction, long bytes, const struct timeval *t0,
               const struct timeval *t1)
{
  struct timeval  td;
  float           s, bs;

  if (ftp_verbose)
  {
    tvsub(&td, t1, t0);
    s = td.tv_sec + (td.tv_usec / 1000000.);
#define nz(x) ((x) == 0 ? 1 : (x))
    bs = bytes / nz(s);
    /* LogPrint(, LOG_LEVEL_INFO, "%ld bytes %s in %.3g secs (%.2g Kbytes/sec)",
     * bytes, direction, s, bs / 1024.0); */
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "%ld bytes %s\n", bytes, direction);
  }
}

void tvsub(struct timeval *tdiff, const struct timeval *t1, const struct timeval *t0)
{

  tdiff->tv_sec = t1->tv_sec - t0->tv_sec;
  tdiff->tv_usec = t1->tv_usec - t0->tv_usec;
  if (tdiff->tv_usec < 0)
    tdiff->tv_sec--, tdiff->tv_usec += 1000000;
}

void lostpeer(int ignore)
{
  //  (void)ignore;

  if (connected)
  {
    if (control_fd != -1)
    {
      //      shutdown(control_fd, 1+1);
      shutdown(control_fd, SHUT_RDWR);
      close(control_fd);
      control_fd = -1;
    }
    if (data >= 0)
    {
      shutdown(data, SHUT_RDWR);
      close(data);
      data = -1;
    }
    connected = 0;
  }
}

t_log          *ftpLogGet(void)
{
  return &ftpLog;
}

#if 0
static int empty(fd_set * mask, int hifd, int sec)
{
  struct timeval  t;

  t.tv_sec = (long)sec;
  t.tv_usec = 0;
  return (select(hifd + 1, mask, (fd_set *) 0, (fd_set *) 0, &t));
}

static void abort_remote(int data_fd)
{
  char            buf[BUFSIZ];
  char            send_buffer[10];
  int             sent = 0;
  int             nfnd, hifd;
  fd_set          mask;
  int             loop = 0;

  /*
   * send IAC in urgent mode instead of DM because 4.3BSD places oob mark
   * after urgent byte rather than before as is protocol now
   */
  printf("abort_remote: aborting");
  sprintf(buf, "%c%c%c", IAC, IP, IAC);
  if (send(control_fd, buf, 3, MSG_OOB) != 3)
    perror("abort");
  sprintf(send_buffer, "%cABORT\r\n", DM);
  if ((sent = write(control_fd, send_buffer, 8)) != 8)
  {
    perror("abort_remote: write error");
  }
  FD_ZERO(&mask);
  FD_SET(control_fd, &mask);
  hifd = control_fd;
  if (data)
  {
    FD_SET(data_fd, &mask);
    if (hifd < data_fd)
      hifd = data;
  }
  if ((nfnd = empty(&mask, hifd, 10)) <= 0)
  {
    if (nfnd < 0)
    {
      perror("abort");
    }
    if (ptabflg)
      code = -1;
    lostpeer(0);
  }
  if ((data_fd != -1) && FD_ISSET(data_fd, &mask))
  {
    printf("abort_remote: looping\n");
    fflush(stdout);
    while (read(data_fd, buf, BUFSIZ) > 0)
    {
      printf("abort_remote: reading %d\n", loop++);
      fflush(stdout);
      /* to detect the server closing the cnx */
      /* LOOP */ ;
    }
  }
  if (getreply(0) == FTP_ERROR && code == 552)
  {
    /* 552 needed for nic style abort */
    printf("abort_remote: second getreply\n");
    fflush(stdout);
    (void)getreply(0);
  }
  (void)getreply(0);
}
#endif

/**
 * @}
**/

/******************************************************************************/
