
/* API of the FTP client */

/**
 * @addtogroup ftp
 * @{
 * */

/**
 * @file ftp_pub.h
 * @brief Implements File Transfer Protocol
 *
 * @author Copyright (c) 1985, 1989 Regents of the University of California.All rights reserved.
 *
 * @note From: @(#)ftp.c  5.38 (Berkeley) 4/22/91
 *
 * */

#ifndef __FTPAPI
#  define __FTPAPI

typedef enum
{
  FTP_ERR_OK = 0,
  FTP_ERR_INVALID_ARGUMENT = -1,
  FTP_ERR_ALREADY_CONNECTED = -2,
  FTP_ERR_CONNECTION_FAILED = -3,
  FTP_ERR_AUTH_FAILED = -4,
  FTP_ERR_FILE_ERROR = -5,
  FTP_ERR_DOWNLOAD_ERROR = -6,
  FTP_ERR_NOT_SUPPORTED = -7
} t_ftpError;

/**
 * 
 * @brief Opens an FTP session with the remote host using the specified account parameters
 *      
 * @param host IP address of the FTP server
 * @param login Login account for this session
 * @param passwd Passwd for the login account
 *
 * @return Return code : 0 on success,
 *   	                   -1 invalid argument(null argument?),
 * 		                   -2 already connected,
 * 		                   -3 failed to connect to remote FTP server,
 * 	                     -4 authentication failed.
 **/
t_ftpError      FtpOpen(char *host, char *login, char *passwd);

/**
 *
 * @brief Closes the current FTP session 
 *
 * @return always 0  
 *
 * */

t_ftpError      FtpClose(void);

/**
 *
 * @brief  Downloads a file in the current FTP session
 *
 * @param remote_name Name of the file to be downloaded from	the remote FTP server
 * @param local_name The name of the file to be set on the local file system
 *
 * @return 0 on success,
 *         1 on failure.
 *
 * */
t_ftpError      FtpGetFile(char *remote_name, char *local_name);

/**
 *
 * @brief Uploads a file in the current FTP session 
 *
 * @param	local_name 	Name of the local file to be uploaded
 * @param remote_name Name of the file to be set on the remote FTP server file system
 * @param append The write mode on the remote server (0 is creation, 1 is append)
 *
 * @return 0 on success,
 *         1 on failure.
 * */
t_ftpError      FtpPutFile(char *local_name, char *remote_name, int append);

/**
 *
 * @brief Downloads a file in the current FTP session	(low level function)
 *
 * @param	remote_name Name of the remote file to be downloaded
 * @param cons The callback to be called when data are read from the FTP data connection 
 * @param bufsize The maximum bytes that can be passed to the callback at once 
 *
 * @return 0 on success,
 *         1 on failure.
 * */
t_ftpError      FtpGet(char *remote_name, int (*cons) (char *, unsigned), unsigned bufsize);

/**
 * @brief Uploads a file in the current FTP session	(low level function)
 *
 * @param remote_name Name of the local file to be uploaded
 * @param	prod The callback to be called to fetch the local data to be sent
 * @param	bufsize The maximum amount of data that can be fetched at once
 * @param	append The write mode on the remote server (0 is creation, 1 is append)
 *
 * @return 0 on success,
 *         1 on failure.
 * */

t_ftpError      FtpPut(char *remote_name, int (*prod) (char *, unsigned), unsigned bufsize,
                       int append);

/**
 * @brief List the files present on the remote server (low level function)
 *
 * @param long_ls ls or ls -l command
 * @param	file Name of the local file to list
 * @param	cons The callback to be called to fetch the local data to be sent
 *
 * @return 0 on success,
 *         1 on failure.
 * */
t_ftpError      FtpLs(int long_ls, char *file, int (*cons) (char *, unsigned));
#endif

/**
 * @}
 * */
