
/**
 * @addtogroup ftp
 * @{
 * */

/**@file ftp_task.c
 * @brief Initialization code
 *
 * @author Thierry Quervel
 * @date 24/03/2003
 *
 * */

/*
 *******************************************************************************
 ** Includes
 *******************************************************************************
 */

#include <types.h>
#include "stdlib.h"

#include <sys/select.h>

#include <netdb.h>

#include "log_pub.h"

#include "ftp_task.h"

#include "ftp_pub.h"
#include "ftp_var.h"
#include "list_pub.h"
#include "console_pub.h"

#define MYBUFSIZ 256

/*
 *******************************************************************************
 ** Local types
 *******************************************************************************
 */

#define MYBUFSIZ 256

/*
 *******************************************************************************
 ** Local types
 *******************************************************************************
 */

/*
 *******************************************************************************
 ** Local variable
 *******************************************************************************
 */

/*
 *******************************************************************************
 ** Static Functions prototypes
 *******************************************************************************
 */

static t_consoleMenu ftpMenu;
static t_consoleCommand ftpOpenCommand;
static t_consoleCommand ftpCloseCommand;
static t_consoleCommand ftpGetCommand;
static t_consoleCommand ftpPutCommand;
static t_consoleCommand ftpPrintCommand;
static t_consoleCommand ftpLsCommand;
static t_consoleCommand ftpLogCommand;

/*
 *******************************************************************************
 ** Function Implementation
 *******************************************************************************
 */

/**
 * @brief Private function to open a FTP connection from the telnet console 
 *
 * @param argum The command to execute
 *
 * @return  CONSOLE_RES_COMMAND_EXEC_OK if the ftp connection is opened. 
 *          CONSOLE_RES_COMMAND_EXEC_KO on failure.
 *
 * */

static t_consoleExecResult FtpOpenExec(char **argum)
{
  int             rc = 0;

  if ((argum[0] == NULL) || (argum[1] == NULL) || (argum[2] == NULL))
  {
    LogPrint(0, 0, "FtpGetExec: Erroneous number of arguments");
    return CONSOLE_RES_COMMAND_EXEC_KO;
  }
  LogPrint(0, 0, "Openning %s@%s:%s\n", argum[0], argum[1], argum[2]);
  if ((rc = FtpOpen(argum[0], argum[1], argum[2])) != 0)
  {
    LogPrint(0, 0, "FtpGetExec: FtpOpen returned error %d\n", rc);
    return CONSOLE_RES_COMMAND_EXEC_KO;
  }
  return CONSOLE_RES_COMMAND_EXEC_OK;
}

/**
 * @brief Private function to close a FTP connection from the telnet console 
 *
 * @param argum The command to execute
 *
 * @return  CONSOLE_RES_COMMAND_EXEC_OK if the ftp connection is closed. 
 *          CONSOLE_RES_COMMAND_EXEC_KO on failure.
 *
 * */
static t_consoleExecResult FtpCloseExec(char **argum)
{
  FtpClose();
  return CONSOLE_RES_COMMAND_EXEC_OK;
}

t_consoleExecResult FtpGetExec(char **argum)
{
  int             err;

  if ((argum[0] == NULL) || (argum[1] == NULL))
  {
    LogPrint(0, 0, "FtpGetExec: Erroneous number of arguments");
    return CONSOLE_RES_COMMAND_EXEC_KO;
  }
  if ((err = FtpGetFile(argum[0], argum[1])) != 0)
  {
    LogPrint(0, 0, "FtpGetExec: FtpGetFile returned error %d", err);
    return CONSOLE_RES_COMMAND_EXEC_KO;
  }
  return CONSOLE_RES_COMMAND_EXEC_OK;
}

/**
 * @brief Private function to transfer a file to a server from the telnet console 
 *
 * @param argum The command to execute
 *
 * @return  CONSOLE_RES_COMMAND_EXEC_OK if the file is transferred. 
 *          CONSOLE_RES_COMMAND_EXEC_KO on failure.
 *
 * */
static t_consoleExecResult FtpPutExec(char **argum)
{
  int             err;

  if ((argum[0] == NULL) || (argum[1] == NULL))
  {
    LogPrint(0, 0, "FtpGetExec: Erroneous number of arguments");
    return CONSOLE_RES_COMMAND_EXEC_KO;
  }
  if ((err = FtpPutFile(argum[0], argum[1], 0)) != 0)
  {
    LogPrint(0, 0, "FtpPutExec: FtpPutFile returned error %d", err);
    return CONSOLE_RES_COMMAND_EXEC_KO;
  }
  return CONSOLE_RES_COMMAND_EXEC_OK;
}

/**
 * @brief The callback used to print dwonloaded files through the console
 *
 * @param buf The buffer
 * @param size The size of the buffer
 *
 * @return Always 0 
 *
 * */
static int print_get_buf(char *buf, unsigned size)
{
  char            mybuf[MYBUFSIZ + 1];
  INT32U          lsize = size;

  if (lsize > MYBUFSIZ)
    lsize = MYBUFSIZ;
  strncpy(mybuf, buf, lsize);
  mybuf[lsize] = '\0';
  //mybuf[lsize] = '\0';
  iprintf("print_get_buf: %s\n", mybuf);
  return 0;
}

/**
 * @brief Private function to print the files in a server from the telnet console 
 *
 * @param argum The command to execute
 *
 * @return  CONSOLE_RES_COMMAND_EXEC_OK if the files are printed. 
 *          CONSOLE_RES_COMMAND_EXEC_KO on failure.
 *
 * */

static t_consoleExecResult FtpPrintExec(char **argum)
{
  unsigned        bufsize = MYBUFSIZ;
  int             err;

  if (argum[0] == NULL)
  {
    LogPrint(0, 0, "FtpGetExec: Erroneous number of arguments");
    return CONSOLE_RES_COMMAND_EXEC_KO;
  }
  if ((err = FtpGet(argum[0], print_get_buf, bufsize)) != 0)
  {
    LogPrint(0, 0, "FtpPrintExec: FtpGet returned error %d", err);
    return CONSOLE_RES_COMMAND_EXEC_KO;
  }
  return CONSOLE_RES_COMMAND_EXEC_OK;
}

/**
 * @brief Private function to print the files in a server from the telnet console 
 *
 * @param argum The command to execute
 *
 * @return  CONSOLE_RES_COMMAND_EXEC_OK if the files are printed. 
 *          CONSOLE_RES_COMMAND_EXEC_KO on failure.
 *
 * */
static t_consoleExecResult FtpLsExec(char **argum)
{

  /*  if(argum[0] == NULL) 
   * {
   * LogPrint(0, 0, "FtpGetExec: Erroneous number of arguments");
   * return CONSOLE_RES_COMMAND_EXEC_KO;
   * }
   */

  int             err;

  if ((err = FtpLs(1, argum[0], print_get_buf)) != 0)
  {
    LogPrint(0, 0, "FtpLsExec: FtpLs returned error %d", err);
    return CONSOLE_RES_COMMAND_EXEC_KO;
  }
  return CONSOLE_RES_COMMAND_EXEC_OK;
}

/**
 * @brief Private function to change the ftp log 
 *
 * @param argum The command to execute
 *
 * @return  CONSOLE_RES_COMMAND_EXEC_OK if the log level aer changed. 
 *          CONSOLE_RES_COMMAND_EXEC_KO on failure.
 *
 * */
static t_consoleExecResult FtpLogExec(char **cmd)
{
  INT32U          logLevel;

  if (cmd[0][0] == 'u')         // unset
  {
    logLevel = strtoul(cmd[1], NULL, 16);
    iprintf("Unset Level 0x%08X\n", logLevel);
    LogLevelUnset(ftpLogGet(), logLevel);
  }
  if (cmd[0][0] == 's')         // set
  {
    logLevel = strtoul(cmd[1], NULL, 16);
    iprintf("Set Level 0x%08X\n", logLevel);
    LogLevelSet(ftpLogGet(), logLevel);
  }

  return CONSOLE_RES_COMMAND_EXEC_OK;
}

void FtpTask(void)
{
  // Configure the console for ftp testing
  // Create the FTP top menu in mode ADMIN
  ConsoleMenuInit(&ftpMenu, "ftp", "ftp", "FTP commands", CONSOLE_MODE_ADMIN);

  // Create open command
  ConsoleCommandInit(&ftpOpenCommand, "open", "open", "Open a FTP connection (host, login, passwd)",
                     CONSOLE_MODE_ADMIN, FtpOpenExec);

  // Create close command
  ConsoleCommandInit(&ftpCloseCommand, "close", "close", "Close the current FTP connection",
                     CONSOLE_MODE_ADMIN, FtpCloseExec);

  // Create getfile command
  ConsoleCommandInit(&ftpGetCommand, "get", "get",
                     "Get a file from a remote server (remote name, local name)",
                     CONSOLE_MODE_ADMIN, FtpGetExec);

  // Create putfile command
  ConsoleCommandInit(&ftpPutCommand, "put", "put",
                     "Put a file onto a remote server (local, remote, append mode)",
                     CONSOLE_MODE_ADMIN, FtpPutExec);

  // Create print command
  ConsoleCommandInit(&ftpPrintCommand, "print", "print",
                     "Print a file from a remote server (remote)", CONSOLE_MODE_ADMIN,
                     FtpPrintExec);

  // Create ls command
  ConsoleCommandInit(&ftpLsCommand, "list", "ll", "list remote files (file name or nothing for *)",
                     CONSOLE_MODE_ADMIN, FtpLsExec);

  // Create log command
  ConsoleCommandInit(&ftpLogCommand, "log", "l", "Log level (set:l s <level>, unset:l u <level>",
                     CONSOLE_MODE_ADMIN, FtpLogExec);

  // Add commands to ftp menu 
  ConsoleCommandAdd(&ftpMenu, &ftpOpenCommand);
  ConsoleCommandAdd(&ftpMenu, &ftpCloseCommand);
  ConsoleCommandAdd(&ftpMenu, &ftpGetCommand);
  ConsoleCommandAdd(&ftpMenu, &ftpPutCommand);
  ConsoleCommandAdd(&ftpMenu, &ftpPrintCommand);
  ConsoleCommandAdd(&ftpMenu, &ftpLsCommand);
  ConsoleCommandAdd(&ftpMenu, &ftpLogCommand);

  ConsoleMenuAdd(ConsoleRootMenuGet(), &ftpMenu);

  LogPrint(0, 0, "\nFTP: CONSOLE INITIALIZATION ENDS\n");

  LogAdd(ftpLogGet());
}

INT8U FTPComponentIsPresent()
{
  return 1;
}

/**
 * @}
 * */
