
/**
 * @addtogroup ftp
 * @{
 * */

/**
 * @file ftplib.h
 * @brief File Transfer Protocol definitions
 *
 * @author Copyright (c) 1985, 1989 Regents of the University of California.All rights reserved.
 *
 * @note From: @(#)ftp.c  5.38 (Berkeley) 4/22/91
 *
 * */

/* for off_t */
#include <sys/types.h>

/// Global parameters. They can be changed at any time between API calls

int             ftp_safeput = 1;  ///< if set, put to destfile".tmp", then rename
int             ftp_verbose = 1;
int             ftp_debug = 1;

/// These default values are what you want. Believe me.
int             ftp_passive = 0;
int             ftp_sendport = 1;

/* Other FTP "parameters" that you can't change:
 * struct, mode and format are obsolete
 * type (ascii/binary) is automatic
 */

/// Basic API

/**
 * @brief Basic API function to open a ftp connection
 *
 * @param host Server Host name
 * @param login user name for host login
 * @param passwd password for host login
 *
 * @return 0 Successful
 *
 * */
int             ftp_open(char *host, char *login, char *passwd);

/**
 * @brief Basic API function to close a ftp connection
 *
 * @return 0 Successful. 1 Means already not connected 
 *
 * */
int             ftp_close(void);

#if 0
int             ftp_cd(char *directory);
int             ftp_mkdir(char *directory);
int             ftp_rmdir(char *directory);
int             ftp_delete(char *filename);
#endif /* 0 */

/**
 * @brief Basic API function to rename a ftp connection
 *
 * @param oldname Old string name 
 * @param newname New String name
 *
 * @return FTP Reply Code.
 *
 * */
int             ftp_rename(char *oldname, char *newname);

/**
 * @brief Basic API function that list the files present on the remote server (low level function)
 *
 * @param	long_ls If 1 "LIST" command requested . If 0 "NLST". 
 * @param file_or_dir File name or directory to list 
 * @param	cons The callback to be called to fetch the local data to be sent
 *
 * @return 0 on success,
 *         1 on failure. 
 * */
int             ftp_ls(int long_ls, char *file_or_dir, int (*cons) (char *, unsigned));

/**
 * @brief High level file transfer functions to get a file
 *
 * @param	local_name 	Name of the local file to be uploaded
 * @param remote_name Name of the file to be get from the remote FTP server file system
 * @param restart_point Restart point offset 
 *
 * @return 0 on success,
 *         1 on failure.
 * */
int             ftp_getfile(char *remote_name, char *local_name, off_t restart_point);

/**
 * @brief High level file transfer functions to put a file
 *
 * @param local_name Name of the local file to be uploaded
 * @param remote_name Name of the file to be set on the remote FTP server file system
 * @param restart_point Restart point offset
 * @param append if 1 -> Append command, if 0 -> Store command
 *
 * @return 0 on success,
 *         1 on failure.
 *
 * */
int             ftp_putfile(char *local_name, char *remote_name, off_t restart_point, int append);

#if 0

/* 'remote_name' is the suggested name to be used by the server
 * 'uname' is a buffer of size 'usize' that will contain on output the
 * unique name generated by the server.
 * if can be set to NULL if you don't care about this name
 */
int             ftp_putfileunique(char *local_name, char *remote_name, char *uname, unsigned usize);
#endif /* 0 */

#if 0

/* send raw FTP commands
 * answer string of the server is put in 'buffer', whose size is 'size'
 * size is limited to BUFSIZ (but it won't be that long anyway)
 */
int             ftp_sendcommand(char *cmd, char *buffer, unsigned size);
#endif /* 0 */

/* low level file transfer functions
 * look at how ftp_getfile(), ftp_putfile(), and ftp_putfileunique()
 * are implemented in ftplib.c to see how to use them
 */

/**
 * @brief Low level file transfer function to get a file
 *
 * @param remote_name Name of the file to be set on the remote FTP server file system
 * @param cons Callback 
 * @param bufsize Buffer size
 * @param restart_point Restart point offset
 *
 * @return int
 *
 * */
int             ftp_get(char *remote_name, int (*cons) (char *, unsigned), unsigned bufsize,
                        off_t restart_point);

/**
 * @brief Low level file transfer function to put a file
 *
 * @param remote_name Name of the file to be set on the remote FTP server file system
 * @param prod callback
 * @param bufsize Buffer size
 * @param restart_point Restart point offset
 * @param append if 1 -> Append command, if 0 -> Store command
 *
 * @return int
 *
 * */
int             ftp_put(char *remote_name, int (*prod) (char *, unsigned), unsigned bufsize,
                        off_t restart_point, int append);

#if 0
int             ftp_putunique(char *remote_name, char *uname, unsigned usize,
                              int (*prod) (char *, unsigned), unsigned bufsize);
#endif /* 0 */

/**
 * @}
 * */
