
/*
 * Copyright 2007 Design of Systems on Silicon (DS2) All rights reserved.
 */

/**
 * @addtogroup sha2
 * @{
**/

/**
 * @file sha2_protocol_pub.h
 * @brief SHA2 message-digest algorithm header file.
 *
 * @internal
 *
 * @author Christophe Devine
 * @date 00/00/2003
 * 
**/

#ifndef SHA2_PROTOCOL_PUB_H
#  define SHA2_PROTOCOL_PUB_H

#  ifdef __cplusplus
extern          "C"
{
#  endif

/**
 * @brief          SHA-256 context structure
 */
  typedef struct
  {
    unsigned long   total[2];   /*!< number of bytes processed  */
    unsigned long   state[8];   /*!< intermediate digest state  */
    unsigned char   buffer[64]; /*!< data block being processed */
  }
  sha2_context;

/**
 * @brief          SHA-256 context setup
 *
 * @param ctx      SHA-256 context to be initialized
 */
  void            sha2_starts(sha2_context * ctx);

/**
 * @brief          SHA-256 process buffer
 *
 * @param ctx      SHA-256 context
 * @param input    buffer holding the  data
 * @param ilen     length of the input data
 */
  void            sha2_update(sha2_context * ctx, unsigned char *input, int ilen);

/**
 * @brief          SHA-256 final digest
 *
 * @param ctx      SHA-256 context
 * @param output   SHA-256 checksum result
 */
  void            sha2_finish(sha2_context * ctx, unsigned char output[32]);

#  ifdef SHA2_PROTOCOL_TEST     // Code not used in FW

/**
 * @brief          Output = SHA-256( input buffer )
 *
 * @param input    buffer holding the  data
 * @param ilen     length of the input data
 * @param output   SHA-256 checksum result
 */
  void            sha2_csum(unsigned char *input, int ilen, unsigned char output[32]);

/**
 * @brief          Output = SHA-256( file contents )
 *
 * @param path     input file name
 * @param output   SHA-256 checksum result
 * @return         0 if successful, or 1 if fopen failed
 */
  int             sha2_file(char *path, unsigned char output[32]);

/**
 * @brief          Output = HMAC-SHA-256( input buffer, hmac key )
 *
 * @param key      HMAC secret key
 * @param keylen   length of the HMAC key
 * @param input    buffer holding the  data
 * @param ilen     length of the input data
 * @param output   HMAC-SHA-256 result
 */
  void            sha2_hmac(unsigned char *key, int keylen, unsigned char *input, int ilen,
                            unsigned char output[32]);

/**
 * @brief          Checkup routine
 *
 * @return         0 if successful, or 1 if the test failed
 */
  int             sha2_self_test(void);
#  endif                        // Code not used in FW

#  ifdef __cplusplus
}
#  endif

#endif                          /* sha2_protocol_pub.h */

/**
 * @}
**/
